import numpy as np
import random as rand
import matplotlib.pyplot as plt	
import time, copy, os

pacfrac=float(os.getcwd().split('/')[-1])/100.
rand.seed(5)


def Area(p):
        #compute the area of a shape    
        return 0.5 * abs(sum(x0*y1 - x1*y0 for ((x0, y0), (x1, y1)) in Segments(p)))

def Segments(p):
        return zip(p, p[1:] + [p[0]])

def TriangleLattice(norder=2,pacfrac=0.87):
	import numpy as np
	import random as rand
	shape=[[0,0],[np.cos(60/180.*np.pi), np.sin(60/180.*np.pi)],[1,0]]
	shapeup=list(shape)
	shapedown=[[0,0],[1,0],[np.cos(60/180.*np.pi), -np.sin(60/180.*np.pi)]]
	#norder=2
	#pacfrac=0.87
	nparticles=2*norder**2
	particlearea=Area(shape)
	boxl=(nparticles*particlearea/pacfrac)**(0.5)
	nperdegree=2*norder
	spacing=boxl/(nperdegree)
	import numpy as np
	l=1.
	h=((3.)**.5)/2.
	N=norder
	a=2*N**2
	b=N**2*(2*h+l)
	c=N**2*h*l-2*N**2*particlearea/pacfrac
	s=np.roots([a,b,c])[-1]
	boxl=N*(l+2*s)
	boxh=N*(h+s)
	print 'PACKING FRACTION',nparticles*particlearea/boxl/boxh
	x=[.5*n*l +n*s for n in range(N*2)]
	y=[-round(.5*n)*h -int(.5*n)*s for n in range(N*2)]
	X,Y=np.meshgrid(x,y)
	z=np.zeros((N*2,N*2))
	ioddcount=0
	ievencount=0
	for i in range(N*2):
		for j in range(N*2):
			if i%2!=0:#if i is odd
				if ioddcount%2==0: # and icount is even
					if j%2!=0:
						z[i][j]=1.
				elif ioddcount%2!=0: #icount is odd
					if j%2==0:
						z[i][j]=1.
			elif i%2==0: #if i is iven
				if ievencount%2==0: # and icount is even
					if j%2==0:
						z[i][j]=-1.
				elif ievencount%2!=0: #icount is odd
					if j%2!=0:
						z[i][j]=-1.		
				
		if i%2==0:
			ievencount+=1
		elif i%2!=0:
			ioddcount+=1
	
	posarray=[]
	for i in range(N*2):
		for j in range(N*2):
			if z[i][j]==1:#put in an up traingle
	                        vel=np.array([rand.random()-.5,rand.random()-.5])
	                        omega=rand.random()-.5
				pos=shapeup+np.array([X[i][j],Y[i][j]])
				posarray.append([pos,vel,omega])
			elif z[i][j]==-1:#put in a down triangle
	                        vel=np.array([rand.random()-.5,rand.random()-.5])
	                        omega=rand.random()-.5
				pos=shapedown+np.array([X[i][j],Y[i][j]])
				posarray.append([pos,vel,omega])
			else: continue
	
	box=np.array([boxl,boxh])
	return np.array(posarray),box



def TotalEnergy(Pos):
	a=np.array(Pos)
	angleke=(a[:,2]**2).sum()
	linearke=sum([(a[i][1]**2).sum() for i in range(len(a))])
	return angleke+linearke	

def Area(p):
	#compute the area of a shape	
	return 0.5 * abs(sum(x0*y1 - x1*y0 for ((x0, y0), (x1, y1)) in Segments(p)))

def Segments(p):
	return zip(p, p[1:] + [p[0]])

def LineCross(pos0,pos1,boxl):
	#translate pos0 to be the minimum image particle to pos0
	minimdist=np.round((pos0.mean(axis=0)-pos1.mean(axis=0))/boxl)
	#print minimdist
	pos0-=boxl*minimdist
	#test if any component lines in the shapes pos0 and pos1 intersect
	result=False
	for i,j in [(k,k+1) for k in range(pos0.shape[0]-1)]+[(-1,0)]:
		for l,m in [(n,n+1) for n in range(pos1.shape[0]-1)]+[(-1,0)]:
			if Intersect(pos0[i],pos0[j],pos1[l],pos1[m]):
				result=True
	return result

def Intersect(a,b,c,d):
	return ((d[1]-a[1])*(c[0]-a[0]) >= (c[1]-a[1])*(d[0]-a[0])) != ((d[1]-b[1])*(c[0]-b[0]) >= (c[1]-b[1])*(d[0]-b[0])) and ((c[1]-a[1])*(b[0]-a[0]) >= (b[1]-a[1])*(c[0]-a[0])) != ((d[1]-a[1])*(b[0]-a[0]) >= (b[1]-a[1])*(d[0]-a[0]))


def InsertParticles(shape,nparticles,boxla):
	boxl=boxla[0]
	shape=np.array(shape,dtype=float)
	#center the polygon
	shape=shape-shape.mean(axis=0)
	Pos=None
	while nparticles>0:
			#print nparticles
			#set up random postion, velocity, omega, and orientation
			vel=np.array([rand.random()-.5,rand.random()-.5])
			omega=rand.random()-.5
			center=np.array([(rand.random()-.5)*boxl,(rand.random()-.5)*boxl])
			#rotate the shape
			theta=rand.random()*2*np.pi
			r=np.array([[np.cos(theta),-np.sin(theta)],[np.sin(theta),np.cos(theta)]])
			particle=np.dot(r,shape.T).T
			#translate the shape
			particle+=center
			if (Pos is None and not ((particle>=boxl/2.).any() or (particle<=-boxl/2.).any() )):
				Pos=[[particle,vel,omega]]
				nparticles-=1
			elif Pos is not None:
				if (not np.array([LineCross(particle,posx,boxla) for posx,y,z in Pos]).any() and not ((particle>=boxl/2.).any() or (particle<=-boxl/2.).any() )):
					Pos.append([particle,vel,omega])
					nparticles-=1
	return Pos

def Cycloid(pos,vel,omega,deltat):
	"""
	pos 	(2,N) array of vertex coordinates for the polygon
	vel	(2,1) array of the velocity components for the polygon
	omega	scalar angular velocity for the polygon
	deltat	(NStep,1) array of deltat values to use for propogating cycloid (e.g., deltat = np.linspace(0,nstep,nstep+1)*dt )
	
	Takes in the information about a polygon particle and returns the coordinate positions of it's vertices propogated through time
	How this works is it rotates and translates the particle so that it is moving in the positive x direction, it then propogates 
	parametric equations for the vertex positions through time, then rotates and particle trajectory back to its original position and direction.
	"""
	pos=np.array(pos,dtype=float)
	vel=np.array(vel,dtype=float)
	#psi is the angel of travel
	psi=np.arctan2(vel[1],vel[0])
	#c is the centroid position
	c=pos.mean(axis=0)
	#magv is the magnitude of the velocity
	magv=np.sqrt(np.dot(vel,vel))
	normv=np.array([vel[1],-vel[0]]/magv) #normalized norm vector to velocity vector
	# 'a' is the 'roll' radius for the cycloid parametric equation
	a=magv/omega
	#'floor' is the point upon which the cycloid wheel is resting, from which point the cycloid rolls
	floor=c+normv*a 
	x,y=([],[])
	for r in pos:
		p=r-c #the vector from centroid to vertex point
		#theta is the initial rotation angle fo the cycloid
		theta=np.arctan2(p[1],p[0])-psi
		#X,Y are the cycloid coordinates along the positive x direction
		b=np.sqrt(np.dot(p,p))
		#The parametric equations for a cycloid with a a general angular velocity
		X= magv*deltat+b*np.cos(theta+omega*deltat)
		Y= a+b*np.sin(theta+omega*deltat)
		#x,y are the cycloid coordincates rotated about the origin by psi, and translated by the vector c
		x.append(X*np.cos(psi)-Y*np.sin(psi)+floor[0])
		y.append(X*np.sin(psi)+Y*np.cos(psi)+floor[1])
	return np.array([x,y]).T

def StepToTime(Pos,time):
	deltat=time
	Posy=copy.deepcopy(Pos)
	#bring the system to a specific point in time
	for i,(pos,vel,omega) in enumerate(Posy):
		#psi is the angel of travel
		psi=np.arctan2(vel[1],vel[0])
		#c is the centroid position
		c=pos.mean(axis=0)
		#magv is the magnitude of the velocity
		magv=np.sqrt(np.dot(vel,vel))
		normv=np.array([vel[1],-vel[0]]/magv) #normalized norm vector to velocity vector
		# 'a' is the 'roll' radius for the cycloid parametric equation
		a=magv/omega
		#'floor' is the point upon which the cycloid wheel is resting, from which point the cycloid rolls
		floor=c+normv*a
		x,y=([],[])
		for ii,r in enumerate(pos):
			p=r-c #the vector from centroid to vertex point
			#theta is the initial rotation angle fo the cycloid
			theta=np.arctan2(p[1],p[0])-psi
			#X,Y are the cycloid coordinates along the positive x direction
			b=np.sqrt(np.dot(p,p))
			#The parametric equations for a cycloid with a a general angular velocity
			X= magv*deltat+b*np.cos(theta+omega*deltat)
			Y= a+b*np.sin(theta+omega*deltat)
			#x,y are the cycloid coordincates rotated about the origin by psi, and translated by the vector c
			Posy[i][0][ii,0]=(X*np.cos(psi)-Y*np.sin(psi)+floor[0])#SPEED ME UP
			Posy[i][0][ii,1]=(X*np.sin(psi)+Y*np.cos(psi)+floor[1])
	return Posy




def PropogateToCollision(Pos,time):
	#Pos is a list [Posi,Posj]
	Pos=copy.deepcopy(Pos)
	deltat=time
        for i,(pos,vel,omega) in enumerate(Pos):
                pos=pos.copy()#np.array(pos,dtype=float)
                vel=np.array(vel,dtype=float)
                #psi is the angel of travel
                psi=np.arctan2(vel[1],vel[0])
                #c is the centroid position
                c=pos.mean(axis=0)
                #magv is the magnitude of the velocity
                magv=np.sqrt(np.dot(vel,vel))
                normv=np.array([vel[1],-vel[0]]/magv) #normalized norm vector to velocity vector
                # 'a' is the 'roll' radius for the cycloid parametric equation
                a=magv/omega
                #'floor' is the point upon which the cycloid wheel is resting, from which point the cycloid rolls
                floor=c+normv*a
                x,y=([],[])
                for ii,r in enumerate(pos):
                        p=r-c #the vector from centroid to vertex point
                        #theta is the initial rotation angle fo the cycloid
                        theta=np.arctan2(p[1],p[0])-psi
                        #X,Y are the cycloid coordinates along the positive x direction
                        b=np.sqrt(np.dot(p,p))
                        #The parametric equations for a cycloid with a a general angular velocity
                        X= magv*deltat+b*np.cos(theta+omega*deltat)
                        Y= a+b*np.sin(theta+omega*deltat)
                        #x,y are the cycloid coordincates rotated about the origin by psi, and translated by the vector c
                        pos[ii,0]=(X*np.cos(psi)-Y*np.sin(psi)+floor[0])#SPEED ME UP
                        pos[ii,1]=(X*np.sin(psi)+Y*np.cos(psi)+floor[1])
		Pos[i][0]=pos
        return Pos


def BuildNeighborList(Pos,boxl,cutoffsq=5):
	centers=[pos.mean(axis=0) for pos,vel,omega in Pos]
	neighborlist=[]
	for i,posi in enumerate(centers):
		dist=(MinImage((centers-posi),boxl)**2).sum(axis=1)
		neighbors=np.array((dist<=cutoffsq).nonzero()).flatten()
		neighborlist.append((neighbors[(neighbors!=i)]).tolist())
	return neighborlist

def DrawFrame(Pos,boxl,suffix=''):
	global plt
	xy=[]
	z=[]
	triangles=[]
	i=0
	for q,(pos,vel,omega) in enumerate(Pos):
		pos=MinImageParticle(pos,boxl)
		z.extend([q]*3)
		xy.extend(pos.tolist())
		triangles.extend([[i,i+1,i+2]])
		minim=False
		nboxlist=[np.array([a,b],dtype=float) for a,b in [[0,1],[1,1],[1,0],[1,-1],[0,-1],[-1,-1],[-1,0],[-1,1]]]
		i+=3
		if 1:
			pos=MinImageParticle(pos,boxl)
			for nbox in nboxlist:
				minpos= pos-boxl*nbox
				z.extend([q]*3)
				xy.extend(minpos.tolist())
				triangles.extend([[i,i+1,i+2]])
				i+=3
	xy=np.asarray(xy)
	x=xy[:,0]
	y=xy[:,1]
	triangles=np.asarray(triangles)
	z=np.array(z)
	plt.clf()
	plt.gca().set_aspect('equal')
	ax1 = plt.axes(frameon=True)
	ax1.axes.get_yaxis().set_visible(False)
	ax1.axes.get_xaxis().set_visible(False)
        plt.xlim([-1*boxl[0]/2.,1*boxl[0]/2.])
        plt.ylim([-1*boxl[1]/2.,1*boxl[1]/2.])
        plt.gca().set_aspect('equal')#plt.xlim([-0.34,-0.32])
        #plt.ylim([-0.19,-0.17])
	plt.tripcolor(x, y, triangles, z, shading='faceted')
	plt.savefig('test'+suffix+'.png',dpi=300)
	#plt.draw()#()
	return


def FindCollision(Pos,boxl,deltat,tol=1e-2,Ignore=None,Visualize=False):
	global plt
	a=np.array([0,0],dtype=float)
	normcp=np.array([0,0],dtype=float)
	normcjp=np.array([0,0],dtype=float)
	njchord=np.array([0,0],dtype=float)
	n=np.array([0,0],dtype=float)
	rapn=np.array([0,0],dtype=float)
	rbpn=np.array([0,0],dtype=float)
	Collision=[]
	CollisionList=[]
	neighbors=BuildNeighborList(Pos,boxl)
	jchords=[(m,m+1) for m in range(3-1)]+[(-1,0)]
	if 1:
		i=1
		Pos=StepToTime(Pos,deltat[-1])
		for ii,(posi,veli,omegai) in enumerate(Pos):
			#analyze for impacts with each neighboring particle
			for j in neighbors[ii]:
				posj=Pos[j][0]#Traj[j][i]#<- the neighbor of interest
				if not LineCross(posi,posj,boxl): continue
				#jchords=[(m,m+1) for m in range(posj.shape[0]-1)]+[(0,-1)]#< uncomment for general polygons
				for iii,p in enumerate(posi):
					#for each vertex point in the particle look for impacts with each choord of the neighboring particle
					for k,l in jchords:
						#to determine if there is an impact first screen the state to see if the vectors from the point
						# to the two choord points are in the same direction
						jchord=posj[l]-posj[k]#<intraparticle distance so no minimage
						njchord[:]=[-jchord[1],jchord[0]]
						pk=MinImage(p-posj[k],boxl)
						val1=np.dot(pk,njchord)
						if abs(val1)<=.1:
							pl=MinImage(p-posj[l],boxl)
							if not abs(abs(np.dot(pk,jchord))+abs(np.dot(pl,jchord))+abs(np.dot(pk,njchord))+abs(np.dot(pl,njchord))-1.)<=.2: continue
							val2=np.dot(pl,njchord)
							#since the point p is along the chord of interest in the neighbor, we should verify that the velocity of point p
							#is in fact bringing it towards a collision, if it is not, then we can assume that no collision is imminent.
							#get the distance from the centroid to point p
							cjp=MinImage(p-posj.mean(axis=0),boxl)
							cp=p-posi.mean(axis=0)
							normcp[:]=[cp[1],-cp[0]]#/cpdist #<- normalized norm vector of cp
							normcjp[:]=[cjp[1],-cjp[0]]
							velpi=-Pos[ii][2]*normcp+Pos[ii][1] #<- omega * || p-c || + velocity
							velpj=-Pos[j][2]*normcjp+Pos[j][1]
							vij= velpi-velpj
							velnorm=np.dot(vij,njchord)
							if np.sign(velnorm)==-1: #<if the y component of the rotated velocity at point p is (-)
								#print 'with colliding velocity',velnorm
								#then the point is moving towards the chord and an impact will/has already occured.
								#Now, test to see if this state is pre or post impact by rotating state so that the choord lies along the x axis
								#(with the centroid in the postive y direction), then look at the sign on the y-coordinate of the point p
								#see if the direction of travel of the point will bring it across the chord 
								#(in both directions in time, since at this time point it may already be tunneled
								#check to see if this is an 'opposite wall corner collisoin'
								val3=np.dot(cp,njchord)
								if ((-val1/np.dot(pk,jchord)>=1.72320508075688772 or val2/np.dot(pl,jchord)>=1.72320508075688772) and not (val3<=-.28866 and val3>=-0.57736 )): continue#this checks if the point p lies more than 60 degrees away from a corner
								#print 'and is not a post collisions corner point'
								p1=MinImage(p-posj[k],boxl)
								#we can get a pretty good idea of the time to impact by dividing the  normal distance of of the rotated point
								# to the chord (p1yrotated) by the y component of the rotated velocity velocity (velprotated)
								timetoimpact=-np.dot(p1,njchord)/velnorm
								if timetoimpact>0: continue
								impacttime1=timetoimpact
								Posx=StepToTime(Pos,timetoimpact)
								p2=Posx[ii][0][iii]
								jchord=Posx[j][0][l]-Posx[j][0][k]#<intraparticle distance so no minimage
								njchord[:]=[-jchord[1],jchord[0]]
								p1=MinImage(p2-Posx[j][0][k],boxl)
								cjp=MinImage(p2-Posx[j][0].mean(axis=0),boxl)
								cp=p2-Posx[ii][0].mean(axis=0)
								normcp[:]=[cp[1],-cp[0]]#/cpdist #<- normalized norm vector of cp
								normcjp[:]=[cjp[1],-cjp[0]]
								velpi=-Posx[ii][2]*normcp+Posx[ii][1] #<- omega * || p-c || + velocity
								velpj=-Posx[j][2]*normcjp+Posx[j][1]
								vij= velpi-velpj
								velnorm=np.dot(vij,njchord)
								timetoimpact=-np.dot(p1,njchord)/velnorm
								impacttime1+=timetoimpact
								if abs(impacttime1)>deltat[1]: continue #this is a resting contact that will not really collide for a very long time so ignore iti                                        
								CollisionList.append([ii,j,iii])
								Collision.append([impacttime1,ii,j,iii,l,k,i,n,vij,normcp,normcjp])
	if len(Collision)>=1:
		impacttime,ii,j,iii,l,k,stepnum,n,vij,normcp,normcjp=sorted(Collision)[0]
		Pos=StepToTime(Pos,impacttime)
		n1=Pos[j][0][l]-Pos[j][0][k]
		n[:]=[-n1[1],n1[0]]
		p=Pos[ii][0][iii]
		rap=p-Pos[ii][0].mean(axis=0)
		rbp=MinImage(p-Pos[j][0].mean(axis=0),boxl)
		rapn[:]=[rap[1],-rap[0]]
		rbpn[:]=[rbp[1],-rbp[0]]
		vap=-Pos[ii][2]*rapn+Pos[ii][1]
		vbp=-Pos[j][2]*rbpn+Pos[j][1]
		impulse=np.dot(-2*(vap-vbp),n)/(2.+(np.dot(rapn,n)**2+np.dot(rbpn,n)**2))
		jn=impulse*n
		Pos[ii][1]+=jn
		Pos[j][1]-=jn
		Pos[ii][2]-=np.dot(rapn,jn)
		Pos[j][2]+=np.dot(rbpn,jn)
		return Pos,deltat[1]+impacttime
		
	else:
		return Pos,deltat[1]
								

def MinImageParticle(particle,boxl):
	#does minimum imaging on a per-particle basis, using the centroid as the imaging point (used for visualizing whole triangles)
	return particle-boxl*np.round(particle.mean(axis=0)/boxl)

def MinImage(vec,boxl):
	#return vec
	return vec-boxl*np.round(vec/boxl)

def GetImpactInfo(posi,vi,wi,posj,vj,wj,thistime):
	return

def ProcessTraj(Traj,boxl):
	NewTraj=[]
	nframe=len(Traj[0])
	for i in range(nframe):
		thisframe=[]
		for particle in Traj:
			thisparticle=MinImageParticle(particle[i],boxl)
			thisframe.extend(thisparticle.tolist())
		NewTraj.append(thisframe)
	return NewTraj

def MakeTriangles(nparticles):
	triangles=[]
	j=0
	for i in range(nparticles):
		triangles.extend([[j,j+1,j+2]])
		j+=3

	return np.asarray(triangles)

def Animate(Traj,Save=False,Visualize=1):
	nparticles=len(Traj)
	triangles=MakeTriangles(nparticles)
	NewTraj=ProcessTraj(Traj)
	if Visualize:
		plt.ion()
		plt.show()
	else:
		plt.figure()
	for i,xy in enumerate(NewTraj):
		xy=np.asarray(xy)
		x=xy[:,0]
		y=xy[:,1]
		triangles=np.asarray(triangles)
		z=np.ones(len(x))
		plt.clf()
		plt.gca().set_aspect('equal')
		plt.tripcolor(x, y, triangles, z, shading='faceted')
		plt.draw()
		if Save:
			plt.savefig('test'+str(i)+'.png')



nstep=1
dt=0.01
deltat = np.linspace(0,nstep,nstep+1)*dt

import pickle

Pos,boxl=TriangleLattice(norder=8,pacfrac=pacfrac)
BuildNeighborList(Pos,boxl)
start=time.clock()
plt.ion()
plt.show()
ax1 = plt.axes(frameon=False)
PosTraj=[]
simtime=0
for q in range(10000000):
	print q,simtime,'\n'
	#start=time.clock()
	Pos,dtime=FindCollision(Pos,deltat=deltat,boxl=boxl,Visualize=0)
	simtime+=dtime
	#print time.clock()-start
	if q%10==0:
		DrawFrame(Pos,boxl,suffix='%8.8d'%q)
	if q%100==0:
		PosTraj.append([simtime,Pos])
		pickle.dump(PosTraj,open('traj.dat','w'))


exit()
#Data Analysis
import pickle
import numpy as np
import matplotlib.pyplot as plt
fig=plt.figure()
ax=plt.subplot(111)
for s in ['42','56','69','74']:
	print 'parsing data',s
	data=np.load(open(s+'traj.dat','r'))
	#parse data
	dataset=[]
	for i,frame in enumerate(data):
		vel=frame[1][0][1]
		pos=frame[1][0][0]
		for site in frame[1][1:]:
			pos=np.vstack((pos,site[0]))
			vel=np.vstack((vel,site[1]))
		if i==0:
			velavg=vel.mean(axis=0)
			print velavg
		dataset.append([frame[0],pos])
	pos0=dataset[len(dataset)/4][1]
	time0=data[len(dataset)/4][0]
	msdtraj=[]
	for time,pos in dataset[len(dataset)/4:]:
		elapsedtime=time-time0
		displacement=sum(sum(((pos-np.array([ 0.01248907,  0.05216564])*elapsedtime)-pos0)**2))/float(pos.shape[0])
		print elapsedtime, displacement
		msdtraj.append([time-time0+1,displacement])
	z=np.array(msdtraj)
	ax.loglog(z[:,0],z[:,1],label=r'$\theta$'+' = .'+s)
	ax.set_xlim(1,1000)
	ax.set_ylim(1,10000)


leg=ax.legend(loc=0)
leg.get_frame().set_linewidth(0)
leg.get_frame().set_alpha(0.5)
for t in leg.get_texts():
	t.set_fontsize('xx-large')

ax.set_xlabel('time',size='xx-large')
ax.set_ylabel('mean-squared displacement',size="xx-large")
fig.subplots_adjust(left=0.12)
fig.subplots_adjust(bottom=0.12)
for tick in ax.xaxis.get_major_ticks():
	tick.label1.set_fontsize(18)

for tick in ax.yaxis.get_major_ticks():
	tick.label1.set_fontsize(18)

for l in ax.get_xticklines():
	l.set_markersize(6)

for l in ax.get_yticklines():
	l.set_markersize(6)

plt.savefig('msd.png',dpi=300)



